/* otfview.c -- View glyphs of OpenType fonts.

Copyright (C) 2003, 2004
  National Institute of Advanced Industrial Science and Technology (AIST)
  Registration Number H15PRO167

This file is part of libotf.

Libotf is free software; you can redistribute it and/or modify it
under the terms of the GNU Lesser General Public License as published
by the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

Libotf is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library, in a file named COPYING; if not,
write to the Free Software Foundation, Inc., 59 Temple Place, Suite
330, Boston, MA 02111-1307, USA.  */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>
#include <libgen.h>

#include "config.h"

#ifdef HAVE_X11_XAW_COMMAND_H

#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Viewport.h>

#include <ft2build.h>
#include FT_FREETYPE_H

#include <otf.h>

#define DEFAULT_PIXEL_SIZE 30
int pixel_size;

#define DEFAULT_FONT_NAME "6x13"
XFontStruct *font;
#define FONT_HEIGHT (font->ascent + font->descent)
#define FONT_ASCENT (font->ascent)
#define FONT_DESCENT (font->descent)
#define FONT_WIDTH (font->max_bounds.width)

XtAppContext context;
/* Widget structure.
   +--- frame (form) -------------------------+
   | +--- command_area (box) ---------------+ |
   | | quit dump charmap ...                | |
   | +--------------------------------------+ |
   | +---- navi_area (box) -----------------+ |
   | | FIRST PREV prev range next NEXT LAST | |
   | +--------------------------------------+ |
   | +--- glyph_area (form) ----------------+ |
   | | idxh[0] glyph[0]    ...    glyph[15] | |
   | |   ...                        ...     | |
   | | idxh[7] glyph[112]  ...    glyph[127]| |
   | |         idxl[0]     ...    idxl[15]  | |
   | +--------------------------------------+ |
   | +--- render_area (form) ---------------+ |
   | | clear del bidi alt_subst             | |
   | | +--- raw (box) --------------------+ | |
   | | | raw_label raw_image              | | |
   | | +----------------------------------+ | |
   | | GSUB all none features...            | |
   | | GPOS all none features...            | |
   | | +--- seq (box) --------------------+ | |
   | | | seq_label seq_image              | | |
   | | +----------------------------------+ | |
   | +--------------------------------------+ |
   +------------------------------------------+ */
Widget shell, frame;
Widget command_area, quit, dump, *charmap;
Widget navi_area, FIRST, PREV, prev, range, next, NEXT, LAST;
Widget glyph_area, glyph[128], index_label[8];
Widget render_area, clear, del, bidi, alt_subst, raw, seq;
Widget raw_label, raw_image, seq_label, seq_image;
unsigned long foreground, background;

typedef struct
{
  OTF_Tag tag;
  int on;
  Widget w;
} FeatureElement;

typedef struct
{
  char *label;
  OTF_GSUB_GPOS *gsub_gpos;
  OTF_LangSys *langsys;
  int num_features;
  FeatureElement *features;
  Widget parent;
} FeatureRec;

FeatureRec gsub, gpos;

OTF_Tag script_tag;

int glyph_char[128];

Display *display;
GC gc, gc_set, gc_or, gc_inv;

typedef struct {
  Pixmap pixmap;
  unsigned width, height;
  int x, y;
  int advance;
} BitmapRec;

BitmapRec bitmap[0x10000];

int render_width, render_height;
Pixmap raw_pixmap, seq_pixmap, gsub_pixmap, gpos_pixmap;
Pixmap none_pixmap;

FT_Face face;

struct {
  int platform_id;
  int encoding_id;
  char name[20];
} charmap_rec[10];

int charmap_index;

int reversed;
int do_alternate_subst;
unsigned glyph_width, glyph_height;
int glyph_x, glyph_y;
int glyph_index;

struct {
  int n_glyphs;
  int glyphs[64];
  int codes[64];
} glyph_rec;

OTF *otf;
char *filename;

void
create_pixmap (int index)
{
  int err = FT_Load_Glyph (face, index, FT_LOAD_RENDER | FT_LOAD_MONOCHROME);
  XImage ximage;
  Pixmap pixmap;
  
  if (err)
    {
      bitmap[index].pixmap = none_pixmap;
      return;
    }
  ximage.height = face->glyph->bitmap.rows;
  ximage.width = face->glyph->bitmap.width;
  ximage.depth = 1;
  ximage.bits_per_pixel = 1;
  ximage.xoffset = 0;
  ximage.format = XYPixmap;
  ximage.data = (char *) face->glyph->bitmap.buffer;
  ximage.byte_order = MSBFirst;
  ximage.bitmap_unit = 8;
  ximage.bitmap_bit_order = MSBFirst;
  ximage.bitmap_pad = 8;
  ximage.bytes_per_line = face->glyph->bitmap.pitch;
  XInitImage (&ximage);
  pixmap = XCreatePixmap (display, DefaultRootWindow (display),
			  glyph_width, glyph_height, 1);
  XFillRectangle (display, pixmap, gc, 0, 0, glyph_width, glyph_height);
  XPutImage (display, pixmap, gc, &ximage, 0, 0,
	     glyph_x + face->glyph->bitmap_left,
	     glyph_y - face->glyph->bitmap_top,
	     ximage.width, ximage.height);
  bitmap[index].pixmap = pixmap;
  bitmap[index].width = ximage.width;
  bitmap[index].height = ximage.height;
  bitmap[index].x = face->glyph->bitmap_left;
  bitmap[index].y = - face->glyph->bitmap_top;
  bitmap[index].advance = face->glyph->metrics.horiAdvance >> 6;
}

void
update_glyph_area ()
{
  int i;
  Arg arg[2];
  char buf[16];
  int msb;

  for (i = 0; i < 128; i++)
    {
      int index = glyph_index + i;

      if (charmap_index >= 0)
	index = FT_Get_Char_Index (face, (FT_ULong) index);
      if (! bitmap[index].pixmap)
	create_pixmap (index);
      XtSetArg (arg[0], XtNbitmap, bitmap[index].pixmap);
      XtSetValues (glyph[i], arg, 1);
    }

  msb = (glyph_index >> 7) % 2 ? 8 : 0;
  for (i = 0; i < 8; i++)
    {
      char str[3];

      sprintf (str, "%XX", i | msb );
      XtSetArg (arg[0], XtNheight, glyph_height + 5);
      XtSetArg (arg[1], XtNlabel, str);
      XtSetValues (index_label[i], arg, 2);
    }

  sprintf (buf, " %04X-%04X ", glyph_index, glyph_index + 0x7F);
  XtSetArg (arg[0], XtNlabel, buf);
  XtSetValues (range, arg, 1);
}

char *
get_features (OTF_FeatureList *list, FeatureRec *rec)
{
  int i, n;
  char *str, *p;

  if (! rec->features[0].on)
    return NULL;
  for (i = n = 0; i < rec->langsys->FeatureCount; i++)
    {
      if (rec->features[i].on)
	n++;
      else
	break;
    }
  if (i == rec->langsys->FeatureCount)
    {
      str = malloc (2);
      strcpy (str, "*");
      return str;
    }
  str = malloc (n * 5);
  for (i = 0, p = str; i < n; i++, p += 5)
    {
      OTF_tag_name (rec->features[i].tag, p);
      p[4] = ',';
    }
  p[-1] = '\0';
  return str;
}


#define DEVICE_DELTA(table, size)				\
  (((size) >= (table).StartSize && (size) <= (table).EndSize)	\
   ? (table).DeltaValue[(size) >= (table).StartSize]		\
   : 0)

void
adjust_anchor (OTF_Anchor *anchor, FT_Face ft_face,
	       OTF_Glyph *g, int *x, int *y)
{
  if (anchor->AnchorFormat == 2)
    {
      FT_Outline *outline;
      int ap = anchor->f.f1.AnchorPoint;

      FT_Load_Glyph (ft_face, (FT_UInt) g->glyph_id, FT_LOAD_MONOCHROME);
      outline = &ft_face->glyph->outline;
      if (ap < outline->n_points)
	{
	  *x = outline->points[ap].x;
	  *y = outline->points[ap].y;
	}
    }
  else if (anchor->AnchorFormat == 3)
    {
      *x += DEVICE_DELTA (anchor->f.f2.XDeviceTable, pixel_size);
      *y += DEVICE_DELTA (anchor->f.f2.YDeviceTable, pixel_size);
    }
}

void
update_seq_area ()
{
  int i, x;
  OTF_GlyphString gstring;
  OTF_Glyph *g, *prev, *base, *mark;
  int base_width;
  int len = glyph_rec.n_glyphs;
  Arg arg[1];
  int unitsPerEm = face->units_per_EM;

  gstring.size = gstring.used = len;
  gstring.glyphs = malloc (sizeof (OTF_Glyph) * len);
  memset (gstring.glyphs, 0, sizeof (OTF_Glyph) * len);
  for (i = 0; i < len; i++)
    gstring.glyphs[i].c = gstring.glyphs[i].glyph_id = glyph_rec.glyphs[i];

  XFillRectangle (display, seq_pixmap, gc, 0, 0, render_width, render_height);
  XDrawLine (display, seq_pixmap, gc_set, 0, glyph_y, render_width, glyph_y);
  if (otf)
    {
      char *str;

      OTF_drive_gdef (otf, &gstring);
      if (otf->gsub)
	{
	  str = get_features (&otf->gsub->FeatureList, &gsub);
	  if (str)
	    {
	      if (do_alternate_subst)
		OTF_drive_gsub_alternate (otf, &gstring, NULL, NULL, str);
	      else
		OTF_drive_gsub (otf, &gstring, NULL, NULL, str);
	      free (str);
	    }
	}
      if (otf->gpos)
	{
	  str = get_features (&otf->gpos->FeatureList, &gpos);
	  if (str)
	    {
	      OTF_drive_gpos (otf, &gstring, NULL, NULL, str);
	      free (str);
	    }
	}
    }

  prev = NULL;
  if (reversed)
    {
      OTF_Glyph temp;

      for (prev = gstring.glyphs, g = gstring.glyphs + gstring.used - 1;
	   prev < g; prev++, g--)
	temp = *prev, *prev = *g, *g = temp;
      for (g = gstring.glyphs; g < gstring.glyphs + gstring.used; g++)
	if (g->GlyphClass == 3)
	  {
	    OTF_Glyph *g0;
	    prev = g++;
	    while (g < gstring.glyphs + gstring.used && g->GlyphClass == 3)
	      g++;
	    for (g0 = g; prev < g0; prev++, g0--)
	      temp = *prev, *prev = *g, *g = temp;
	  }
    }


  mark = base = NULL;
  for (i = 0, x = glyph_x, prev = NULL, g = gstring.glyphs;
       i < gstring.used; i++, prev = g++)
    {
      BitmapRec *bmp = bitmap + gstring.glyphs[i].glyph_id;
      int xoff = 0, yoff = 0;
      int prev_width;
      int advance = bmp->advance;
      int j;

      if (! bmp->pixmap)
	{
	  create_pixmap (gstring.glyphs[i].glyph_id);
	  if (! bmp->pixmap)
	    continue;
	  advance = bmp->advance;
	}
      switch (g->positioning_type)
	{
	case 0:
	  break;

	case 1: case 2:
	  {
	    int format = g->f.f1.format;

	    if (format & OTF_XPlacement)
	      xoff = g->f.f1.value->XPlacement * pixel_size / unitsPerEm;
	    if (format & OTF_XPlaDevice)
	      xoff += DEVICE_DELTA (g->f.f1.value->XPlaDevice, pixel_size);
	    if (format & OTF_YPlacement)
	      yoff = g->f.f1.value->YPlacement * pixel_size / unitsPerEm;
	    if (format & OTF_YPlaDevice)
	      yoff += DEVICE_DELTA (g->f.f1.value->YPlaDevice, pixel_size);
	    if (format & OTF_XAdvance)
	      advance += g->f.f1.value->XAdvance * pixel_size / unitsPerEm;
	    if (format & OTF_XAdvDevice)
	      advance += DEVICE_DELTA (g->f.f1.value->XAdvDevice, pixel_size);
	  }
	  break;

	case 3:
	  /* Not yet supported.  */
	  break;
	case 4: case 5:
	  if (! base)
	    break;
	  prev = base;
	  prev_width = base_width;
	  goto label_adjust_anchor;
	default: 		/* i.e. case 6 */
	  if (! mark)
	    break;
	  prev = mark;
	  prev_width = 0;
	label_adjust_anchor:
	  {
	    int base_x, base_y, mark_x, mark_y;

	    base_x = g->f.f4.base_anchor->XCoordinate * pixel_size / unitsPerEm;
	    base_y = g->f.f4.base_anchor->YCoordinate * pixel_size / unitsPerEm;
	    mark_x = g->f.f4.mark_anchor->XCoordinate * pixel_size / unitsPerEm;
	    mark_y = g->f.f4.mark_anchor->YCoordinate * pixel_size / unitsPerEm;

	    if (g->f.f4.base_anchor->AnchorFormat != 1)
	      adjust_anchor (g->f.f4.base_anchor, face, prev, &base_x, &base_y);
	    if (g->f.f4.mark_anchor->AnchorFormat != 1)
	      adjust_anchor (g->f.f4.mark_anchor, face, g, &mark_x, &mark_y);
	    xoff = (base_x - prev_width) - mark_x;
	    yoff = base_y - mark_y;
	  }
	}
	  
      XCopyArea (display, bmp->pixmap, seq_pixmap, gc_or,
		 glyph_x + bmp->x, glyph_y + bmp->y, bmp->width, bmp->height,
		 x + bmp->x + xoff, glyph_y + bmp->y - yoff);
      x += advance;

      if (g->GlyphClass == OTF_GlyphClass0)
	base = mark = g, base_width = advance;
      else if (g->GlyphClass == OTF_GlyphClassMark)
	mark = g;
      else
	base = g, base_width = advance;
    }
  free (gstring.glyphs);

  XtSetArg (arg[0], XtNbitmap, seq_pixmap);
  XtSetValues (seq_image, arg, 1);
}


void
update_render_area ()
{
  int i;
  int x;
  Arg arg[1];

  XFillRectangle (display, raw_pixmap, gc, 0, 0, render_width, render_height);
  for (i = 0, x = glyph_x; i < glyph_rec.n_glyphs; i++)
    {
      BitmapRec *bmp = bitmap + glyph_rec.glyphs[i];
      char buf[5];

      XCopyArea (display, bmp->pixmap, raw_pixmap, gc,
		 0, 0, glyph_width, glyph_height,
		 (glyph_width + 4) * i + 1, 1);
      XDrawRectangle (display, raw_pixmap, gc_set,
		      (glyph_width + 4) * i, 0,
		      glyph_width + 1, glyph_height + 1);
      XDrawLine (display, raw_pixmap, gc_set,
		 (glyph_width + 4) * i + 1 + glyph_x, 1,
		 (glyph_width + 4) * i + 1 + glyph_x, glyph_height + 1);
      XDrawLine (display, raw_pixmap, gc_set,
		 (glyph_width + 4) * i + 1 + glyph_x + bmp->advance, 1,
		 (glyph_width + 4) * i + 1 + glyph_x + bmp->advance,
		 glyph_height + 1);

      sprintf (buf, "%04X", glyph_rec.codes[i]);
      XDrawString (display, raw_pixmap, gc_inv, 
		   (glyph_width + 1) * i + 1
		   + (glyph_width - XTextWidth (font, buf, 4)) / 2,
		   glyph_height + 2 + FONT_HEIGHT, buf, 4);
    }
  XtSetArg (arg[0], XtNbitmap, raw_pixmap);
  XtSetValues (raw_image, arg, 1);
  update_seq_area ();
}

void
QuitProc (Widget w, XtPointer client_data, XtPointer call_data)
{
  XtAppSetExitFlag (XtWidgetToApplicationContext (w));
}

void
DumpProc (Widget w, XtPointer client_data, XtPointer call_data)
{
  int g_width, g_height, g_x, g_y, pix_width, pix_height;
  int margin = 20 * 300 / 25.4;
  int a4_width = 210 * 300 / 25.4 - margin * 2;
  int a4_height = 297 * 300 / 25.4 - margin * 2;
  int size = 100;
  Pixmap pixmap;
  XImage ximage, *image;
  int i, x, y;
  char *data;
  int bytes_per_line;

  FT_Set_Pixel_Sizes (face, 0, size);
  g_width = ((face->bbox.xMax - face->bbox.xMin) * size / face->units_per_EM);
  g_height = ((face->bbox.yMax - face->bbox.yMin) * size / face->units_per_EM);
  pix_width = (g_width + 1) * 16 + margin + 1;
  pix_height = (g_height + 1 + FONT_HEIGHT) * 16 + margin + 1;
  while (pix_width > a4_width || pix_height > a4_height)
    {
      size--;
      FT_Set_Pixel_Sizes (face, 0, size);
      g_width = ((face->bbox.xMax - face->bbox.xMin)
		 * size / face->units_per_EM);
      g_height = ((face->bbox.yMax - face->bbox.yMin)
		  * size / face->units_per_EM);
      pix_width = (g_width + 1) * 16 + margin + 1;
      pix_height = (g_height + 1 + FONT_HEIGHT) * 16 + margin + 1;
    }

  g_x = - (face->bbox.xMin * size / face->units_per_EM);
  g_y = face->bbox.yMax * size / face->units_per_EM;
  for (i = 0; i < 0xFF; i++)
    {
      int idx;

      if (charmap_index >= 0)
	idx = FT_Get_Char_Index (face, (FT_ULong) i);
      else
	idx = i;
      if (FT_Load_Glyph (face, idx, FT_LOAD_RENDER | FT_LOAD_MONOCHROME) == 0)
	{
	  if (g_x < - face->glyph->bitmap_left)
	    g_x = - face->glyph->bitmap_left;
	  if (g_y < face->glyph->bitmap_top)
	    g_y = face->glyph->bitmap_top;
	  if (g_width
	      < g_x + face->glyph->bitmap_left + face->glyph->bitmap.width)
	    g_width
	      = g_x + face->glyph->bitmap_left + face->glyph->bitmap.width;
	  if (g_height
	      < g_y - face->glyph->bitmap_top + face->glyph->bitmap.rows)
	    g_height
	      = g_y - face->glyph->bitmap_top + face->glyph->bitmap.rows;
	}
    }
  pix_width = (g_width + 1) * 16 + margin + 1;
  pix_height = (g_height + FONT_HEIGHT + 1) * 16 + margin + 1;
  pixmap = XCreatePixmap (display,
			  RootWindow (display, DefaultScreen (display)),
			  pix_width, pix_height, 1);
  XFillRectangle (display, pixmap, gc, 0, 0, pix_width, pix_height);

  for (i = 0, x = margin; i <= 16; i++, x += g_width + 1)
    XDrawLine (display, pixmap, gc_set, x, margin,
	       x, margin + (g_height + FONT_HEIGHT + 1) * 16);
  for (i = 0, y = margin; i <= 16; i++, y += g_height + FONT_HEIGHT + 1)
    XDrawLine (display, pixmap, gc_set, margin, y,
	       margin + (g_width + 1) * 16, y);
  for (i = 0; i < 256; i++)
    {
      char str[5];
      int idx;

      if (charmap_index >= 0)
	idx = FT_Get_Char_Index (face, (FT_ULong) i);
      else
	idx = i;
      x = margin + (g_width + 1) * (i % 16);
      y = margin + (g_height + FONT_HEIGHT + 1) * (i / 16);
      if (FT_Load_Glyph (face, idx, FT_LOAD_RENDER | FT_LOAD_MONOCHROME) == 0)
	{
	  ximage.height = face->glyph->bitmap.rows;
	  ximage.width = face->glyph->bitmap.width;
	  ximage.depth = 1;
	  ximage.bits_per_pixel = 1;
	  ximage.xoffset = 0;
	  ximage.format = XYPixmap;
	  ximage.data = (char *) face->glyph->bitmap.buffer;
	  ximage.byte_order = MSBFirst;
	  ximage.bitmap_unit = 8;
	  ximage.bitmap_bit_order = MSBFirst;
	  ximage.bitmap_pad = 8;
	  ximage.bytes_per_line = face->glyph->bitmap.pitch;
	  XInitImage (&ximage);
	  XPutImage (display, pixmap, gc, &ximage, 0, 0,
		     x + g_x + face->glyph->bitmap_left,
		     y + g_y - face->glyph->bitmap_top, 
		     ximage.width, ximage.height);
	}
      sprintf (str, "0x%02X", i);
      XDrawString (display, pixmap, gc_inv,
		   x + (g_width - XTextWidth (font, str, 4))/ 2,
		   y + g_height + FONT_ASCENT, str, 4);
    }

  image = XGetImage (display, pixmap, 0, 0, pix_width, pix_height,
		     AllPlanes, XYPixmap);
  XInitImage (image);
  {
    char *name = alloca (strlen (filename) + 5);
    FILE *fp;

    sprintf (name, "%s.pbm", filename);
    printf ("Writing %s ...", name);
    fp = fopen (name, "w");
    fprintf (fp, "P4\n%d %d\n", image->width, image->height);
    bytes_per_line = (image->width + 7) / 8;
    data = image->data;
    for (y = 0; y < image->height; y++, data += image->bytes_per_line)
      fwrite (data, 1, bytes_per_line, fp);
    fclose (fp);
    printf ("done\n");
  }
  FT_Set_Pixel_Sizes (face, 0, (int) pixel_size);
}


void
GlyphProc (Widget w, XtPointer client_data, XtPointer call_data)
{
  int old_glyph_index = glyph_index;

  if ((int) client_data == -3 && glyph_index > 0)
    glyph_index = 0;
  else if ((int) client_data == -2 && glyph_index > 0)
    glyph_index = (glyph_index - 1) & 0xF000;
  else if ((int) client_data == -1 && glyph_index > 0)
    glyph_index -= 0x80;
  else if ((int) client_data == 1 && glyph_index < 0xFF80)
    glyph_index += 0x80;
  else if ((int) client_data == 2 && glyph_index < 0xF000)
    glyph_index = (glyph_index + 0x1000) & 0xF000;
  else if ((int) client_data == 3 && glyph_index < 0xF000)
    glyph_index = 0xFF80;
  if (glyph_index != old_glyph_index)
    update_glyph_area ();
}

void
CharmapProc (Widget w, XtPointer client_data, XtPointer call_data)
{
  if (charmap_index == (int) client_data)
    return;
  charmap_index = (int) client_data;
  if (charmap_index >= 0)
    FT_Set_Charmap (face, face->charmaps[charmap_index]);
  update_glyph_area ();
}

void
RenderProc (Widget w, XtPointer client_data, XtPointer call_data)
{
  if ((int) client_data < 0)
    {
      if (glyph_rec.n_glyphs > 0)
	{
	  if ((int) client_data == -2)
	    glyph_rec.n_glyphs--;
	  else
	    glyph_rec.n_glyphs = 0;
	  update_render_area ();
	}
    }
  else if (glyph_rec.n_glyphs < 64)
    {
      int index = glyph_index + (int) client_data;

      if (charmap_index >= 0)
	index = FT_Get_Char_Index (face, (FT_ULong) index);
      if (bitmap[index].pixmap)
	{
	  glyph_rec.codes[glyph_rec.n_glyphs] = glyph_index + (int) client_data;
	  glyph_rec.glyphs[glyph_rec.n_glyphs++] = index;
	  update_render_area ();
	}
    }
}

void
BidiProc (Widget w, XtPointer client_data, XtPointer call_data)
{
  Arg arg[1];

  reversed = ! reversed;
  if (reversed)
    XtSetArg (arg[0], XtNlabel, "L<-R");
  else
    XtSetArg (arg[0], XtNlabel, "L->R");
  XtSetValues (w, arg, 1);
  update_seq_area ();
}

void
AltSubstProc (Widget w, XtPointer client_data, XtPointer call_data)
{
  do_alternate_subst = ! do_alternate_subst;
  update_seq_area ();
}

void
FeatureProc (Widget w, XtPointer client_data, XtPointer call_data)
{
  FeatureRec *rec = (FeatureRec *) client_data;
  int idx, i, j;
  Arg arg[4];
  char *label;

  if (! rec->langsys)
    return;
  XtSetArg (arg[0], XtNlabel, &label);
  XtGetValues (w, arg, 1);
  if (! strcmp (label, "all"))
    idx = -2;
  else if (! strcmp (label, "none"))
    idx = -1;
  else
    {
      for (idx = 0; idx < rec->langsys->FeatureCount; idx++)
	if (rec->features[idx].w == w)
	  break;
      if (idx == rec->langsys->FeatureCount)
	idx = -1;
    }
  if (idx < 0)
    {
      int on = idx == -2;

      for (i = j = 0; j < rec->langsys->FeatureCount; j++)
	{
	  int index  = rec->langsys->FeatureIndex[j];

	  rec->features[j].tag
	    = rec->gsub_gpos->FeatureList.Feature[index].FeatureTag;
	  rec->features[j].on = on;
	}
    }
  else
    {
      OTF_Tag tag = rec->features[idx].tag;

      i = idx;
      if (rec->features[i].on)
	{
	  for (j = i + 1;
	       j < rec->langsys->FeatureCount && rec->features[j].on; j++)
	    rec->features[j - 1].tag = rec->features[j].tag;
	  rec->features[j - 1].tag = tag;
	  rec->features[j - 1].on = 0;
	}
      else
	{
	  for (j = i + 1; i > 0 && ! rec->features[i - 1].on; i--)
	    rec->features[i].tag = rec->features[i - 1].tag;
	  rec->features[i].tag = tag;
	  rec->features[i].on = 1;
	}
    }

  for (; i < j; i++)
    {
      char str[5];

      if (rec->features[i].on)
	{
	  XtSetArg (arg[0], XtNborderWidth, 3);
	  XtSetArg (arg[1], XtNinternalHeight, 2);
	  XtSetArg (arg[2], XtNinternalWidth, 2);
	}
      else
	{
	  XtSetArg (arg[0], XtNborderWidth, 1);
	  XtSetArg (arg[1], XtNinternalHeight, 4);
	  XtSetArg (arg[2], XtNinternalWidth, 4);
	}
      OTF_tag_name (rec->features[i].tag, str);
      XtSetArg (arg[3], XtNlabel, str);
      XtSetValues (rec->features[i].w, arg, 4);
    }
  update_seq_area ();
}

void
setup_feature_rec (FeatureRec *rec)
{
  int i;
  Arg arg[10];

  rec->langsys = NULL;
  for (i = 0; i < rec->gsub_gpos->ScriptList.ScriptCount; i++)
    if (rec->gsub_gpos->ScriptList.Script[i].ScriptTag == script_tag)
      {
	rec->langsys = &rec->gsub_gpos->ScriptList.Script[i].DefaultLangSys;
	break;
      }

  if (! rec->features)
    {
      rec->num_features = rec->langsys->FeatureCount;
      rec->features = malloc ((sizeof (FeatureElement)) * rec->num_features);
      memset (rec->features, 0, (sizeof (FeatureElement)) * rec->num_features);
    }
  else if (rec->num_features < rec->langsys->FeatureCount)
    {
      rec->features = realloc (rec->features, 
			       (sizeof (FeatureElement))
			       * rec->langsys->FeatureCount);
      memset (rec->features + rec->num_features, 0,
	      (sizeof (FeatureElement)) 
	      * (rec->langsys->FeatureCount - rec->num_features));
      rec->num_features = rec->langsys->FeatureCount;
    }

  if (! rec->langsys)
    i = 0;
  else
    {
      XtSetArg (arg[0], XtNborderWidth, 1);
      XtSetArg (arg[1], XtNinternalHeight, 4);
      XtSetArg (arg[2], XtNinternalWidth, 4);
      XtSetArg (arg[3], XtNborderColor, foreground);
      XtSetArg (arg[4], XtNsensitive, True);
      for (i = 0; i < rec->langsys->FeatureCount; i++)
	{
	  OTF_Feature *feature = rec->gsub_gpos->FeatureList.Feature;
	  int index = rec->langsys->FeatureIndex[i];
	  char label[5];

	  if (! rec->features[i].w)
	    {
	      Widget w = XtCreateManagedWidget ("", commandWidgetClass,
						rec->parent, arg, 0);
	      XtAddCallback (w, XtNcallback, FeatureProc, (XtPointer) rec);
	      rec->features[i].w = w;
	    }

	  rec->features[i].tag = feature[index].FeatureTag;
	  rec->features[i].on = 0;
	  OTF_tag_name (rec->features[i].tag, label);
	  XtSetArg (arg[5], XtNlabel, label);
	  XtSetValues (rec->features[i].w, arg, 6);
	}
    }
  XtSetArg (arg[0], XtNborderColor, background);
  XtSetArg (arg[1], XtNsensitive, False);
  XtSetArg (arg[2], XtNlabel, "    ");
  for (; i < rec->langsys->FeatureCount; i++)
    {
      if (! rec->features[i].w)
	{
	  Widget w = XtCreateManagedWidget ("", commandWidgetClass,
					    rec->parent, arg, 0);
	  XtAddCallback (w, XtNcallback, FeatureProc, (XtPointer) rec);
	  rec->features[i].w = w;
	}
      XtSetValues (rec->features[i].w, arg, 3);
    }
}

void
ScriptProc (Widget w, XtPointer client_data, XtPointer call_data)
{
  if (script_tag == (OTF_Tag) client_data)
    return;
  script_tag = (OTF_Tag) client_data;
  setup_feature_rec (&gsub);
  setup_feature_rec (&gpos);
  update_seq_area ();
}

Widget
create_otf_script_widgets (Widget prev)
{
  Widget w;
  Arg arg[10];
  int n, i;
  OTF_Tag *scripts;
  char script_name[5];

  XtSetArg (arg[0], XtNborderWidth, 0);
  XtSetArg (arg[1], XtNleft, XawChainLeft);
  XtSetArg (arg[2], XtNright, XawChainLeft);
  XtSetArg (arg[3], XtNtop, XawChainTop);
  XtSetArg (arg[4], XtNbottom, XawChainTop);
  XtSetArg (arg[5], XtNfromVert, prev);
  XtSetArg (arg[6], XtNorientation, XtorientHorizontal);
  prev = XtCreateManagedWidget ("Script", boxWidgetClass, render_area, arg, 7);
  XtCreateManagedWidget ("script", labelWidgetClass, prev, arg, 1);

  n = 0;
  if (otf->gsub)
    n = otf->gsub->ScriptList.ScriptCount;
  if (otf->gpos)
    n += otf->gpos->ScriptList.ScriptCount;
  scripts = alloca (sizeof (OTF_Tag) * n);
  i = 0;
  if (otf->gsub)
    for (; i < otf->gsub->ScriptList.ScriptCount; i++)
      scripts[i] = otf->gsub->ScriptList.Script[i].ScriptTag;
  n = i;
  if (otf->gpos)
    for (; i < otf->gpos->ScriptList.ScriptCount; i++)
      {
	OTF_Tag tag = otf->gpos->ScriptList.Script[i].ScriptTag;
	int j;

	for (j = 0; j < i; j++)
	  if (tag == scripts[j])
	    break;
	if (j == i)
	  scripts[n++] = tag;
      }

  if (n == 0)
    return prev;

  script_tag = scripts[0];
  OTF_tag_name (scripts[0], script_name);
  if (n == 1)
    {
      XtSetArg (arg[0], XtNforeground, background);
      XtSetArg (arg[1], XtNbackground, foreground);
      XtCreateManagedWidget (script_name, labelWidgetClass, prev, arg, 2);
    }
  else
    {
      char name[5];

      XtSetArg (arg[0], XtNstate, True);
      w = XtCreateManagedWidget (script_name, toggleWidgetClass, prev, arg, 1);
      XtAddCallback (w, XtNcallback, ScriptProc, (XtPointer) scripts[0]);
      XtSetArg (arg[0], XtNradioGroup, w);
      for (i = 1; i < n; i++)
	{
	  OTF_tag_name (scripts[i], name);
	  w = XtCreateManagedWidget (name, toggleWidgetClass, prev, arg, 1);
	  XtAddCallback (w, XtNcallback, ScriptProc, (XtPointer) scripts[i]);
	}	  
    }
  return prev;
}


Widget
create_otf_widgets (Widget prev, FeatureRec *rec)
{
  Arg arg[10];
  Widget w;
  int i;

  XtSetArg (arg[0], XtNborderWidth, 0);
  XtSetArg (arg[1], XtNleft, XawChainLeft);
  XtSetArg (arg[2], XtNright, XawChainLeft);
  XtSetArg (arg[3], XtNtop, XawChainTop);
  XtSetArg (arg[4], XtNbottom, XawChainTop);
  XtSetArg (arg[5], XtNfromVert, prev);
  XtSetArg (arg[6], XtNorientation, XtorientHorizontal);
  prev = XtCreateManagedWidget (rec->label, boxWidgetClass, render_area,
				arg, 7);
  XtCreateManagedWidget (rec->label, labelWidgetClass, prev, arg, 1);
  XtSetArg (arg[0], XtNborderWidth, 1);
  XtSetArg (arg[1], XtNinternalHeight, 4);
  XtSetArg (arg[2], XtNinternalWidth, 4);
  w = XtCreateManagedWidget ("all", commandWidgetClass, prev, arg, 3);
  XtAddCallback (w, XtNcallback, FeatureProc, (XtPointer) rec);
  w = XtCreateManagedWidget ("none", commandWidgetClass, prev, arg, 3);
  XtAddCallback (w, XtNcallback, FeatureProc, (XtPointer) rec);

  rec->parent = prev;
  setup_feature_rec (rec);
  return prev;
}

void
create_widgets ()
{
  String quit_action = "<KeyPress>q: set() notify() unset()";
  String FIRST_action = "<KeyPress>f: set() notify() unset()\n\
			 <KeyPress>Home: set() notify() unset()";
  String PREV_action = "Shift<KeyPress>p: set() notify() unset()\n\
			 <KeyPress>Up: set() notify() unset()";
  String prev_action = "~Shift<KeyPress>p: set() notify() unset()\n\
			 <KeyPress>Left: set() notify() unset()";
  String next_action = "~Shift<KeyPress>n: set() notify() unset()\n\
			 <KeyPress>Right: set() notify() unset()";
  String NEXT_action = "Shift<KeyPress>n: set() notify() unset()\n\
			 <KeyPress>Down: set() notify() unset()";
  String LAST_action = "<KeyPress>l: set() notify() unset()\n\
			 <KeyPress>End: set() notify() unset()";
  Arg arg[10];
  int i, j;
  Widget prev, w;
  String trans = "<Expose>: Expose()";

  XtSetArg (arg[0], XtNtranslations, XtParseTranslationTable (trans));
  frame = XtCreateManagedWidget ("frame", formWidgetClass, shell, arg, 1);

  XtSetArg (arg[0], XtNleft, XawChainLeft);
  XtSetArg (arg[1], XtNright, XawChainLeft);
  XtSetArg (arg[2], XtNtop, XawChainTop);
  XtSetArg (arg[3], XtNbottom, XawChainTop);
  XtSetArg (arg[4], XtNborderWidth, 0);
  XtSetArg (arg[5], XtNorientation, XtorientHorizontal);
  command_area = XtCreateManagedWidget ("command-area", boxWidgetClass,
					frame, arg, 6);
  XtSetArg (arg[6], XtNfromVert, command_area);
  navi_area = XtCreateManagedWidget ("navi-area", boxWidgetClass,
				     frame, arg, 7);
  XtSetArg (arg[4], XtNborderWidth, 0);
  XtSetArg (arg[5], XtNfromVert, navi_area);
  XtSetArg (arg[6], XtNdefaultDistance, 0);
  glyph_area = XtCreateManagedWidget ("glyph-area", formWidgetClass,
				      frame, arg, 7);
  XtSetArg (arg[5], XtNfromVert, glyph_area);
  render_area = XtCreateManagedWidget ("render-area", formWidgetClass,
				       frame, arg, 6);

  XtSetArg (arg[0], XtNaccelerators, XtParseAcceleratorTable (quit_action));
  quit = XtCreateManagedWidget ("Quit", commandWidgetClass,
				command_area, arg, 1);
  XtAddCallback (quit, XtNcallback, QuitProc, NULL);

  dump = XtCreateManagedWidget ("Dump Image", commandWidgetClass,
				command_area, arg, 1);
  XtAddCallback (dump, XtNcallback, DumpProc, NULL);

  XtSetArg (arg[0], XtNborderWidth, 0);
  XtSetArg (arg[1], XtNwidth, 10);
  XtCreateManagedWidget ("spacer", boxWidgetClass, command_area, arg, 2);

  charmap = alloca (sizeof (Widget) * (face->num_charmaps + 1));
  XtSetArg (arg[0], XtNstate, True);
  charmap[0] = XtCreateManagedWidget (charmap_rec[0].name, toggleWidgetClass,
				      command_area, arg, 1);
  XtAddCallback (charmap[0], XtNcallback, CharmapProc, (XtPointer) -1);
  XtSetArg (arg[0], XtNradioGroup, charmap[0]);
  for (i = 0; i < face->num_charmaps; i++)
    {
      charmap[i + 1] = XtCreateManagedWidget (charmap_rec[i + 1].name,
					      toggleWidgetClass,
					      command_area, arg, 1);
      XtAddCallback (charmap[i + 1], XtNcallback, CharmapProc, (XtPointer) i);
    }

  XtSetArg (arg[0], XtNlabel, " |< (f)");
  XtSetArg (arg[1], XtNaccelerators, XtParseAcceleratorTable (FIRST_action));
  FIRST = XtCreateManagedWidget ("FIRST", commandWidgetClass,
				 navi_area, arg, 2);
  XtAddCallback (FIRST, XtNcallback, GlyphProc, (XtPointer) -3);
  XtSetArg (arg[0], XtNlabel, "<< (P)");
  XtSetArg (arg[1], XtNaccelerators, XtParseAcceleratorTable (PREV_action));
  PREV = XtCreateManagedWidget ("PREV", commandWidgetClass,
				navi_area, arg, 2);
  XtAddCallback (PREV, XtNcallback, GlyphProc, (XtPointer) -2);
  XtSetArg (arg[0], XtNlabel, "< (p)");
  XtSetArg (arg[1], XtNaccelerators, XtParseAcceleratorTable (prev_action));
  prev = XtCreateManagedWidget ("prev", commandWidgetClass,
				navi_area, arg, 2);
  XtAddCallback (prev, XtNcallback, GlyphProc, (XtPointer) -1);
  XtSetArg (arg[0], XtNlabel, " 0000 ");
  range = XtCreateManagedWidget ("range", labelWidgetClass,
				 navi_area, arg, 1);
  XtSetArg (arg[0], XtNforeground, &foreground);
  XtSetArg (arg[1], XtNbackground, &background);
  XtGetValues (range, arg, 2);

  XtSetArg (arg[0], XtNlabel, "> (n)");
  XtSetArg (arg[1], XtNaccelerators, XtParseAcceleratorTable (next_action));
  next = XtCreateManagedWidget ("next", commandWidgetClass,
				navi_area, arg, 2);
  XtAddCallback (next, XtNcallback, GlyphProc, (XtPointer) 1);
  XtSetArg (arg[0], XtNlabel, ">> (N)");
  XtSetArg (arg[1], XtNaccelerators, XtParseAcceleratorTable (NEXT_action));
  NEXT = XtCreateManagedWidget ("NEXT", commandWidgetClass,
				navi_area, arg, 2);
  XtAddCallback (NEXT, XtNcallback, GlyphProc, (XtPointer) 2);
  XtSetArg (arg[0], XtNlabel, ">| (l)");
  XtSetArg (arg[1], XtNaccelerators, XtParseAcceleratorTable (LAST_action));
  LAST = XtCreateManagedWidget ("LAST", commandWidgetClass,
				navi_area, arg, 2);
  XtAddCallback (LAST, XtNcallback, GlyphProc, (XtPointer) 3);

  XtSetArg (arg[0], XtNleft, XawChainLeft);
  XtSetArg (arg[1], XtNright, XawChainLeft);
  XtSetArg (arg[2], XtNtop, XawChainTop);
  XtSetArg (arg[3], XtNbottom, XawChainTop);

  for (i = 0; i < 8; i++)
    {
      char str[3];
      int n = 4;
      Widget head;

      sprintf (str, "%XX", i);
      XtSetArg (arg[n], XtNheight, glyph_height + 5), n++;
      XtSetArg (arg[n], XtNlabel, str), n++;
      XtSetArg (arg[n], XtNborderWidth, 0), n++;
      if (i > 0)
	XtSetArg (arg[n], XtNfromVert, w), n++;
      head = XtCreateManagedWidget (str, labelWidgetClass, glyph_area, arg, n);
      index_label[i] = head;
      for (j = 0; j < 16; j++)
	{
	  int k = i * 16 + j;

	  n = 4;
	  if (i > 0)
	    XtSetArg (arg[n], XtNfromVert, w), n++;
	  if (j == 0)
	    XtSetArg (arg[n], XtNfromHoriz, head), n++;
	  else
	    XtSetArg (arg[n], XtNfromHoriz, glyph[k - 1]), n++;
	  XtSetArg (arg[n], XtNbitmap, none_pixmap), n++;
	  glyph[k] = XtCreateManagedWidget ("glyph", commandWidgetClass,
					    glyph_area, arg, n);
	  XtAddCallback (glyph[k], XtNcallback, RenderProc, (XtPointer) k);
	}
      w = head;
    }
  /* 10 = (1 (border_width) + 4 (inner_width)) * 2 */
  XtSetArg(arg[4], XtNwidth, glyph_width + 10);
  XtSetArg (arg[5], XtNfromVert, glyph[112]);
  XtSetArg (arg[6], XtNfromHoriz, w);
  XtSetArg (arg[7], XtNborderWidth, 0);

  for (j = 0; j < 16; j++)
    {
      char str[3];

      sprintf (str, "X%X", j);
      XtSetArg (arg[8], XtNlabel, str);
      w = XtCreateManagedWidget ("idx", labelWidgetClass, glyph_area, arg, 9);
      XtSetArg (arg[6], XtNfromHoriz, w);
    }

  XtSetArg (arg[0], XtNleft, XawChainLeft);
  XtSetArg (arg[1], XtNright, XawChainLeft);
  XtSetArg (arg[2], XtNtop, XawChainTop);
  XtSetArg (arg[3], XtNbottom, XawChainTop);
  XtSetArg (arg[4], XtNorientation, XtorientHorizontal);
  XtSetArg (arg[5], XtNborderWidth, 0);
  w = XtCreateManagedWidget ("clear-box", boxWidgetClass, render_area, arg, 6);
  clear = XtCreateManagedWidget ("clear", commandWidgetClass, w, arg, 0);
  XtAddCallback (clear, XtNcallback, RenderProc, (XtPointer) -1);
  del = XtCreateManagedWidget ("delete", commandWidgetClass, w, arg, 0);
  XtAddCallback (del, XtNcallback, RenderProc, (XtPointer) -2);
  bidi = XtCreateManagedWidget ("L->R", toggleWidgetClass, w, arg, 0);
  XtAddCallback (bidi, XtNcallback, BidiProc, NULL);
  alt_subst = XtCreateManagedWidget ("AltSubst", toggleWidgetClass, w, arg, 0);
  XtAddCallback (alt_subst, XtNcallback, AltSubstProc, NULL);

  XtSetArg (arg[4], XtNorientation, XtorientHorizontal);
  XtSetArg (arg[5], XtNborderWidth, 0);
  XtSetArg (arg[6], XtNfromVert, w);
  raw = XtCreateManagedWidget ("raw", boxWidgetClass, render_area, arg, 7);

  XtSetArg (arg[0], XtNborderWidth, 0);
  XtSetArg (arg[1], XtNlabel, "raw: ");
  raw_label = XtCreateManagedWidget ("raw-label", labelWidgetClass,
				      raw, arg, 2);
  XtSetArg (arg[1], XtNbitmap, raw_pixmap);
  raw_image = XtCreateManagedWidget ("raw-image", labelWidgetClass,
				      raw, arg, 2);
  w = raw;
  if (otf)
    {
      OTF_get_table (otf, "GSUB");
      OTF_get_table (otf, "GPOS");
      w = create_otf_script_widgets (w);
      if (otf->gsub)
	{
	  gsub.label = "GSUB";
	  gsub.gsub_gpos = otf->gsub;
	  w = create_otf_widgets (w, &gsub);
	}
      if (otf->gpos)
	{
	  gpos.label = "GPOS";
	  gpos.gsub_gpos = otf->gpos;
	  w = create_otf_widgets (w, &gpos);
	}
    }

  XtSetArg (arg[6], XtNfromVert, w);
  seq = XtCreateManagedWidget ("seq", boxWidgetClass, render_area, arg, 7);
  XtSetArg (arg[0], XtNborderWidth, 0);
  XtSetArg (arg[1], XtNlabel, "seq: ");
  seq_label = XtCreateManagedWidget ("seq-label", labelWidgetClass,
				     seq, arg, 2);
  XtSetArg (arg[1], XtNbitmap, seq_pixmap);
  seq_image = XtCreateManagedWidget ("seq-image", labelWidgetClass,
				     seq, arg, 2);
  XtInstallAllAccelerators (shell, shell);
}

static void
ExposeProc (Widget w, XEvent *event, String *str, Cardinal *num)
{
  XTextProperty text_prop;
  char *pname = "otfview";
  char *fname = basename (filename);
  char *name = alloca (strlen (fname) + 3 + strlen (pname) + 1);

  sprintf (name, "%s - %s", pname, fname);
  text_prop.value = (unsigned char *) name;
  text_prop.encoding = XA_STRING;
  text_prop.format = 8;
  text_prop.nitems = strlen (name);
  XSetWMName (display, XtWindow (shell), &text_prop);
}

/* Format MSG by FMT and print the result to the stderr, and exit.  */

#define FATAL_ERROR(fmt, arg)	\
  do {				\
    fprintf (stderr, fmt, arg);	\
    exit (1);			\
  } while (0)

static int
x_error_handler (Display *display, XErrorEvent *error)
{
  return 0;
}


int
main (int argc, char **argv)
{
  XtActionsRec actions[] = { {"Expose", ExposeProc} };
  Arg arg[10];

  FT_Library library;
  OTF_GlyphString gstring;
  OTF_Glyph *g;

  int err;
  int i;
  int fixed_pixel_size = 0;
  int display_width;

  pixel_size = DEFAULT_PIXEL_SIZE;
  {
    char *str = getenv ("PIXEL_SIZE");

    if (str && (i = atoi (str)) > 0)
      {
	pixel_size = i;
	fixed_pixel_size = 1;
      }
  }

  gstring.size = gstring.used = 256;
  g = calloc (256, sizeof (OTF_Glyph));
  gstring.glyphs = g;

  shell = XtOpenApplication (&context, "OTFView", NULL, 0, &argc, argv, NULL,
			     shellWidgetClass, arg, 0);
  display = XtDisplay (shell);
  /*XSynchronize (display, True);*/
  XSetErrorHandler (x_error_handler);
  display_width = DisplayWidth (display,
				XScreenNumberOfScreen (XtScreen (shell)));
  font = XLoadQueryFont (display, DEFAULT_FONT_NAME);
  if (! font)
    font = XLoadQueryFont (display, "fixed");

  if (argc != 2 || !strcmp (argv[1], "-h") || !strcmp (argv[1], "--help"))
    {
      fprintf (stderr, "Usage: %s [ X-OPTION ... ]  OTF-FILE\n",
	       basename (argv[0]));
      fprintf (stderr,
	       "  Pixel size is decided by the environment variable PIXEL_SIZE ((default %d).\n", DEFAULT_PIXEL_SIZE);
      exit (argc != 2);
    }
  filename = argv[1];
  if (strstr (filename, ".ttf")
      || strstr (filename, ".TTF")
      || strstr (filename, ".otf")
      || strstr (filename, ".OTF"))
    {
      otf = OTF_open (filename);
      if (! otf
	  || OTF_get_table (otf, "head") < 0
	  || OTF_get_table (otf, "cmap") < 0
	  || (OTF_check_table (otf, "GSUB") < 0
	      && OTF_check_table (otf, "GPOS") < 0))
	otf = NULL;
    }

  if ((err = FT_Init_FreeType (&library)))
    FATAL_ERROR ("%s\n", "FT_Init_FreeType: error");
  err = FT_New_Face (library, filename, 0, &face);
  if (err == FT_Err_Unknown_File_Format)
    FATAL_ERROR ("%s\n", "FT_New_Face: unknown file format");
  else if (err)
    FATAL_ERROR ("%s\n", "FT_New_Face: unknown error");
  if ((err = FT_Set_Pixel_Sizes (face, 0, pixel_size)))
    FATAL_ERROR ("%s\n", "FT_Set_Pixel_Sizes: error");

  {
    char title[256];
    Arg arg[1];

    filename = basename (filename);
    sprintf (title, "%s family:%s style:%s",
	     filename, face->family_name, face->style_name);
    XtSetArg (arg[0], XtNtitle, title);
    XtSetValues (shell, arg, 1);
  }

  glyph_width = ((face->bbox.xMax - face->bbox.xMin)
		 * pixel_size / face->units_per_EM);
  if (! fixed_pixel_size && glyph_width * 16 > display_width * 0.8)
    {
      pixel_size = (pixel_size * display_width * 0.8 / 16 / glyph_width);
      FT_Set_Pixel_Sizes (face, 0, pixel_size);
      glyph_width = ((face->bbox.xMax - face->bbox.xMin)
		     * pixel_size / face->units_per_EM);
    }
  if (glyph_width < FONT_WIDTH * 4)
    glyph_width = FONT_WIDTH * 4;

  glyph_height = ((face->bbox.yMax - face->bbox.yMin)
		  *  pixel_size / face->units_per_EM);

  glyph_x = - (face->bbox.xMin * pixel_size / face->units_per_EM);
  glyph_y = face->bbox.yMax * pixel_size / face->units_per_EM;
  none_pixmap = XCreatePixmap (display, DefaultRootWindow (display),
			       glyph_width, glyph_height, 1);

  {
    unsigned long valuemask =  GCFunction | GCLineWidth;
    XGCValues values;

    gc = XCreateGC (display, none_pixmap, (unsigned long) 0, NULL);
    values.function = GXset;
    values.line_width = 1;
    gc_set = XCreateGC (display, none_pixmap, valuemask, &values);
    values.function = GXor;
    gc_or = XCreateGC (display, none_pixmap, valuemask, &values);
    values.function = GXcopyInverted;
    gc_inv = XCreateGC (display, none_pixmap, valuemask, &values);
  }

  XFillRectangle (display, none_pixmap, gc, 0, 0,
		  glyph_width, glyph_height);
  XDrawString (display, none_pixmap, gc_inv,
	       (glyph_width - XTextWidth (font, "none", 4)) / 2,
	       glyph_height / 2, "none", 4);

  render_width = (glyph_width + 4) * 15 + 1;
  render_height = glyph_height + 2;

  charmap_rec[0].platform_id = -1;
  charmap_rec[0].encoding_id = -1;
  strcpy (charmap_rec[0].name, "no charmap");

  for (i = 0; i < face->num_charmaps; i++)
    {
      charmap_rec[i + 1].platform_id = face->charmaps[i]->platform_id;
      charmap_rec[i + 1].encoding_id = face->charmaps[i]->encoding_id;
      sprintf (charmap_rec[i + 1].name, "%d-%d",
	       charmap_rec[i + 1].platform_id, charmap_rec[i + 1].encoding_id);
      if (face->charmaps[i]->platform_id == 0
	  || (face->charmaps[i]->platform_id == 3
	      && face->charmaps[i]->encoding_id == 1))
	strcat (charmap_rec[i + 1].name, " (unicode)");
      else if (face->charmaps[i]->platform_id == 1
	       && face->charmaps[i]->encoding_id == 0)
	strcat (charmap_rec[i + 1].name, " (apple-roman)");
    }

  raw_pixmap = XCreatePixmap (display, DefaultRootWindow (display),
			      render_width, render_height, 1);
  seq_pixmap = XCreatePixmap (display, DefaultRootWindow (display),
			      render_width, render_height, 1);

  memset (bitmap, 0, sizeof (bitmap));
  create_widgets ();
  glyph_index = 0;
  charmap_index = -1;
  update_glyph_area ();
  update_render_area ();

  XtAppAddActions (context, actions, XtNumber (actions));
  XtRealizeWidget (shell);
  XtAppMainLoop (context);

  exit (0);
}

#else /* not HAVE_X11_XAW_COMMAND_H */

int
main (int argc, char **argv)
{
  fprintf (stderr, 
	   "Building of this program failed (lack of some header files)\n",
	   argv[0]);
  exit (1);
}

#endif
